@@ -1,33 +1,20 @@
-#!/usr/bin/python
 """
 HTTP-based ReStructuredText viewer.
-
-Usage:
-    restview [options] filename.rst [...]
-or
-    restview [options] directory [...]
-or
-    restview [options] -e "command" [--watch filename] [...]
-or
-    restview [options] --long-description
-or
-    restview --help
-
-Needs docutils and a web browser. Will syntax-highlight code or doctest blocks
-(needs pygments).
 """
 from __future__ import print_function
 
+import argparse
+import fnmatch
 import os
 import re
-import sys
-import time
 import socket
-import optparse
-import threading
 import subprocess
+import sys
+import threading
+import time
 import webbrowser
 
+
 try:
     import BaseHTTPServer
 except ImportError:
@@ -56,9 +43,8 @@ except ImportError:
 import docutils.core
 import docutils.writers.html4css1
 import pygments
-from pygments import lexers, formatters
-
-from restview.pypi_support import trim_docstring, checkLinkSchemes
+import readme_renderer.rst as readme_rst
+from pygments import formatters, lexers
 
 
 try:
@@ -67,7 +53,7 @@ except NameError:
     unicode = str
 
 
-__version__ = "2.2.1"
+__version__ = '2.9.3'
 
 
 # If restview is ever packaged for Debian, this'll likely be changed to
@@ -89,6 +75,14 @@ class MyRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
         self.do_GET_or_HEAD()
 
     def do_GET_or_HEAD(self):
+        host = self.headers.get('Host', '').split(':', 1)[0]
+        if not any(fnmatch.fnmatch(host, pat)
+                   for pat in self.server.renderer.allowed_hosts):
+            # Protect against DNS rebinding attacks
+            # (https://en.wikipedia.org/wiki/DNS_rebinding)
+            self.log_error("Rejecting unknown Host header: %r", host)
+            self.send_error(400, "Host header not in allowed list")
+            return
         self.path = unquote(self.path)
         if '..' in self.path:
             self.send_error(400, "Bad request") # no hacking!
@@ -128,6 +122,8 @@ class MyRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
             return self.handle_image(self.translate_path(), 'image/png')
         elif self.path.endswith('.jpg') or self.path.endswith('.jpeg'):
             return self.handle_image(self.translate_path(), 'image/jpeg')
+        elif self.path.endswith('.svg'):
+            return self.handle_image(self.translate_path(), 'image/svg+xml')
         elif self.path.endswith('.txt') or self.path.endswith('.rst'):
             return self.handle_rest_file(self.translate_path(), watch)
         else:
@@ -210,7 +206,9 @@ class MyRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
             stdout, stderr = p.communicate()
             if p.returncode != 0:
                 self.log_error("'%s' terminated with %s", command, p.returncode)
-            if stderr or not stdout:
+            if stderr:
+                self.log_error("stderr from '%s':\n%s", command, stderr)
+            if not stdout:
                 return self.handle_error(command, p.returncode, stderr, mtime=mtime)
             else:
                 return self.handle_rest_data(stdout, mtime=mtime)
@@ -382,6 +380,11 @@ pre.error {
     color: red;
     background: #fff8f8;
 }
+pre > .highlight {
+    display: block;
+    color: red;
+    background: #fff8f8;
+}
 </style>
 </head>
 <body>
@@ -408,6 +411,7 @@ class RestViewer(object):
     handler_class = MyRequestHandler
 
     local_address = ('localhost', 0)
+    allowed_hosts = ['localhost', '127.0.0.1']
 
     # Comma-separated list of URLs, full filenames, or filenames in the
     # default search path (if you want to refer to docutils default
@@ -417,7 +421,8 @@ class RestViewer(object):
 
     favicon_path = os.path.join(DATA_PATH, 'favicon.ico')
 
-    strict = False
+    report_level = None
+    halt_level = None
     pypi_strict = False
 
     def __init__(self, root, command=None, watch=None):
@@ -451,49 +456,85 @@ class RestViewer(object):
             writer.translator_class = SyntaxHighlightingHTMLTranslator
         if self.stylesheets:
             stylesheet_dirs = writer.default_stylesheet_dirs + [DATA_PATH]
-            # docutils can't embed http:// or https:// URLs
-            embed_stylesheet = '//' not in self.stylesheets
-            settings_overrides = {'stylesheet': self.stylesheets,
-                                  'stylesheet_path': None,
-                                  'stylesheet_dirs': stylesheet_dirs,
-                                  'embed_stylesheet': embed_stylesheet}
+            if '//' not in self.stylesheets:
+                settings_overrides = {'stylesheet': None,
+                                      'stylesheet_path': self.stylesheets,
+                                      'stylesheet_dirs': stylesheet_dirs,
+                                      'embed_stylesheet': True}
+            else:
+                # docutils can't embed http:// or https:// URLs
+                settings_overrides = {'stylesheet': self.stylesheets,
+                                      'stylesheet_path': None,
+                                      'stylesheet_dirs': stylesheet_dirs,
+                                      'embed_stylesheet': False}
         else:
             settings_overrides = {}
         settings_overrides['syntax_highlight'] = 'short'
         if self.pypi_strict:
-            settings_overrides['raw_enabled'] = 0
-            settings_overrides['file_insertion_enabled'] = 0
-            settings_overrides['halt_level'] = 2
-            rest_input = trim_docstring(rest_input)
-        if self.strict:
-            settings_overrides['halt_level'] = 1
+            settings_overrides.update(readme_rst.SETTINGS)
+        if self.halt_level is not None:
+            settings_overrides['halt_level'] = self.halt_level
+        if self.report_level is not None:
+            settings_overrides['report_level'] = self.report_level
 
-        if settings:
+        if settings:  # hook for unit tests
             settings_overrides.update(settings)
 
         try:
-            if self.pypi_strict:
-                document = docutils.core.publish_doctree(source=rest_input,
-                    source_path=filename,
-                    settings_overrides=settings_overrides)
-                checkLinkSchemes(document)
             docutils.core.publish_string(rest_input, writer=writer,
                                          source_path=filename,
                                          settings_overrides=settings_overrides)
+            if self.pypi_strict:
+                clean_body = readme_rst.clean(''.join(writer.body))
+                if clean_body is None:
+                    # Unfortunately the real error was caught and discared,
+                    # without even logging :/
+                    raise ValueError("Output cleaning failed")
+                writer.body = [clean_body]
+                writer.output = writer.apply_template()
         except Exception as e:
-            html = self.render_exception(e.__class__.__name__, str(e), rest_input, mtime=mtime)
+            line = self.extract_line_info(e, filename)
+            html = self.render_exception(e.__class__.__name__, str(e), rest_input, mtime=mtime, line=line)
         else:
             html = writer.output
         return self.inject_ajax(html, mtime=mtime)
 
-    def render_exception(self, title, error, source, mtime=None):
+    @staticmethod
+    def extract_line_info(exception, source_path):
+        # Docutils constructs a nice system_message object that has
+        # attributes like 'source' and 'line', but then it flattens
+        # that out into a string passed to the exception constructor.
+        msg = str(exception)
+        source_path = str(source_path)  # handle None like docutils
+        if msg.startswith(source_path + ':'):
+            lineno = msg[len(source_path) + 1:].partition(':')[0]
+            if lineno.isdigit():
+                return int(lineno)
+        return None
+
+    @staticmethod
+    def highlight_line(source, lineno):
+        source = escape(source)
+        if not lineno:
+            return source
+        lines = source.splitlines(True)
+        if not 1 <= lineno <= len(lines):
+            return source
+        idx = lineno - 1
+        return ''.join(
+            lines[:idx]
+            + ['<span class="highlight">%s</span>' % lines[idx]]
+            + lines[idx + 1:]
+        )
+
+    def render_exception(self, title, error, source, line=None, mtime=None):
         # NB: source is a bytestring (see issue #16 for the reason)
         # UTF-8 is not necessarily the right thing to use here, but
         # garbage is better than a crash, right?
         source = source.decode('UTF-8', 'replace')
         html = (ERROR_TEMPLATE.replace('$title', escape(title))
                               .replace('$error', escape(error))
-                              .replace('$source', escape(source)))
+                              .replace('$source', self.highlight_line(source, line)))
         return self.inject_ajax(html, mtime=mtime)
 
     def inject_ajax(self, markup, mtime=None):
@@ -503,8 +544,7 @@ class RestViewer(object):
             return markup
 
 
-class SyntaxHighlightingHTMLTranslator(docutils.writers.html4css1.HTMLTranslator):
-
+class SyntaxHighlightingHTMLTranslator(readme_rst.ReadMeHTMLTranslator):
     in_doctest = False
     in_text = False
     in_reference = False
@@ -640,43 +680,72 @@ def launch_browser(url):
 
 
 def main():
-    progname = os.path.basename(sys.argv[0])
-    parser = optparse.OptionParser("%prog [options] filename-or-directory [...]",
+    parser = argparse.ArgumentParser(
+                    usage="%(prog)s [options] root [...]",
                     description="Serve ReStructuredText files over HTTP.",
-                    prog=progname)
-    parser.add_option('-l', '--listen', metavar='PORT',
-                      help='listen on a given port (or interface:port,'
-                           ' e.g. *:8080) [default: random port on localhost]',
-                      default=None)
-    parser.add_option('-b', '--browser',
-                      help='open a web browser [default: only if -l'
-                           ' was not specified]',
-                      action='store_true', default=None)
-    parser.add_option('-e', '--execute', metavar='COMMAND',
-                      help='run a command to produce ReStructuredText',
-                      default=None)
-    parser.add_option('-w', '--watch', metavar='FILENAME', action='append',
-                      help='reload the page when a file changes (use with'
-                           ' --execute); can be specified multiple times',
-                      default=[])
-    parser.add_option('--long-description',
-                      help='run "python setup.py --long-description" to produce'
-                           ' ReStructuredText; also enables --pypi-strict'
-                           ' and watches the usual long description sources'
-                           ' (setup.py, README.rst, CHANGES.rst)',
-                      action='store_true')
-    parser.add_option('--css', metavar='URL|FILENAME',
-                      help='use the specified stylesheet; can be specified'
-                           ' multiple times [default: %s]'
-                           % RestViewer.stylesheets,
-                      action='append', dest='stylesheets', default=[])
-    parser.add_option('--strict',
-                      help='halt at the slightest problem',
-                      action='store_true', default=False)
-    parser.add_option('--pypi-strict',
-                      help='enable additional restrictions that PyPI performs',
-                      action='store_true', default=False)
-    opts, args = parser.parse_args(sys.argv[1:])
+                    prog="restview")
+    parser.add_argument('root',
+                        help='filename or directory to serve documents from',
+                        nargs='*')
+    parser.add_argument('--version', action='version', version=__version__)
+    parser.add_argument('-l', '--listen', metavar='PORT',
+                        help='listen on a given port (or interface:port,'
+                             ' e.g. *:8080) [default: random port on localhost]',
+                        default=None)
+    parser.add_argument('--allowed-hosts', metavar='HOSTS',
+                        help='allowed values for the Host header (default:'
+                             ' localhost only, unless you specify -l *:port,'
+                             ' in which case any Host: is accepted by default)',
+                        default=None)
+    parser.add_argument('-b', '--browser',
+                        help='open a web browser [default: only if -l'
+                             ' was not specified]',
+                        action='store_true', default=None)
+    parser.add_argument('-B', '--no-browser',
+                        help="don't open a web browser",
+                        action='store_false', dest='browser')
+    parser.add_argument('-e', '--execute', metavar='COMMAND',
+                        help='run a command to produce ReStructuredText',
+                        default=None)
+    parser.add_argument('-w', '--watch', metavar='FILENAME', action='append',
+                        help='reload the page when a file changes (use with'
+                             ' --execute); can be specified multiple times',
+                        default=[])
+    parser.add_argument('--long-description',
+                        help='run "python setup.py --long-description" to produce'
+                             ' ReStructuredText; also enables --pypi-strict'
+                             ' and watches the usual long description sources'
+                             ' (setup.py, README.rst, CHANGES.rst)',
+                        action='store_true')
+    parser.add_argument('--css', metavar='URL|FILENAME',
+                        help='use the specified stylesheet; can be specified'
+                             ' multiple times [default: %s]'
+                             % RestViewer.stylesheets,
+                        action='append', dest='stylesheets', default=[])
+    parser.add_argument(
+        '--report-level',
+        help='''set the "report_level" option of docutils; restview
+            will report system messages at or above this level (1=info,
+            2=warnings, 3=errors, 4=severe)''',
+        type=int, default=2)
+    halt_level_group = parser.add_mutually_exclusive_group()
+    halt_level_group.add_argument(
+        '--halt-level',
+        help='''set the "halt_level" option of docutils; restview
+            will stop processing the document when a system message
+            at or above this level (1=info, 2=warnings, 3=errors,
+            4=severe) is logged''',
+        type=int, default=None)
+    halt_level_group.add_argument(
+        '--strict',
+        help='halt at the slightest problem; equivalent to --halt-level=2',
+        action='store_const', dest='halt_level', const=2)
+    parser.add_argument(
+        '--pypi-strict',
+        help='enable additional restrictions that PyPI performs',
+        action='store_true', default=False)
+    opts = parser.parse_args(sys.argv[1:])
+    args = opts.root
     if opts.long_description:
         opts.execute = 'python setup.py --long-description'
         opts.watch += ['setup.py', 'README.rst', 'CHANGES.rst']
@@ -695,7 +764,8 @@ def main():
         server = RestViewer(args, watch=opts.watch)
     if opts.stylesheets:
         server.stylesheets = ','.join(opts.stylesheets)
-    server.strict = opts.strict
+    server.report_level = opts.report_level
+    server.halt_level = opts.halt_level
     server.pypi_strict = opts.pypi_strict
 
     if opts.listen:
@@ -703,6 +773,10 @@ def main():
             server.local_address = parse_address(opts.listen)
         except ValueError as e:
             parser.error(str(e))
+    if opts.allowed_hosts:
+        server.allowed_hosts = opts.allowed_hosts.replace(',', ' ').split()
+    elif server.local_address[0] in ('*', '0', '0.0.0.0'):
+        server.allowed_hosts = ['*']
     host = get_host_name(server.local_address[0])
     port = server.listen()
     try:
@@ -715,8 +789,3 @@ def main():
         pass
     finally:
         server.close()
-
-
-if __name__ == '__main__':
-    main()
-
