@@ -5,15 +5,23 @@ import socket
 import unittest
 import webbrowser
 
+
 try:
     from cStringIO import StringIO
 except ImportError:
     from io import StringIO
 
+import docutils.utils
 from mock import Mock, patch
 
-from restview.restviewhttp import (MyRequestHandler, RestViewer,
-                                   get_host_name, launch_browser, main)
+from restview.restviewhttp import (
+    MyRequestHandler,
+    RestViewer,
+    get_host_name,
+    launch_browser,
+    main,
+)
+
 
 try:
     unicode
@@ -38,24 +46,31 @@ class PopenStub(object):
 
 class MyRequestHandlerForTests(MyRequestHandler):
     def __init__(self):
-        self._headers = {}
+        self.headers = {'Host': 'localhost'}  # request headers
+        self._headers = {}  # response headers
         self.log = []
         self.server = Mock()
         self.server.renderer.command = None
         self.server.renderer.watch = None
+        self.server.renderer.allowed_hosts = ['localhost']
         self.server.renderer.rest_to_html = lambda data, mtime=None, filename=None: \
             unicode('HTML for %s with AJAX poller for %s' % (data, mtime))
         self.server.renderer.render_exception = lambda title, error, source, mtime=None: \
             unicode('HTML for error %s: %s: %s' % (title, error, source))
+
     def send_response(self, status):
         self.status = status
+
     def send_header(self, header, value):
         self._headers[header] = value
+
     def end_headers(self):
         self.headers = self._headers
+
     def send_error(self, status, body):
         self.status = status
         self.error_body = body
+
     def log_error(self, message, *args):
         if args:
             message = message % args
@@ -190,11 +205,21 @@ class TestMyRequestHandler(unittest.TestCase):
         self.assertEqual(handler.status, 400)
         self.assertEqual(handler.error_body, "Bad request")
 
+    def test_do_GET_or_HEAD_prevent_dns_rebinding(self):
+        handler = MyRequestHandlerForTests()
+        handler.headers['Host'] = 'mallory.example.com:80'
+        handler.path = '/'
+        handler.server.renderer.root = self.filepath('a.txt')
+        handler.do_GET_or_HEAD()
+        self.assertEqual(handler.status, 400)
+        self.assertEqual(handler.error_body, "Host header not in allowed list")
+
     def test_do_GET_or_HEAD_images(self):
         for filename, ctype in [('a.png', 'image/png'),
                                 ('a.gif', 'image/gif'),
                                 ('a.jpg', 'image/jpeg'),
                                 ('a.jpeg', 'image/jpeg'),
+                                ('a.svg', 'image/svg+xml'),
                                 ('favicon.ico', 'image/x-icon')]:
             handler = MyRequestHandlerForTests()
             handler.path = '/' + filename
@@ -241,8 +266,10 @@ class TestMyRequestHandler(unittest.TestCase):
         stat = {filename: [Mock(st_mtime=123456)]}
         with patch('os.stat', lambda fn: stat[fn].pop(0)):
             handler.handle_polling([filename], 123455)
-        self.assertEqual(handler.log,
-             ['connection reset by peer (client closed "/polling?pathname=__init__.py&mtime=123455" before acknowledgement)'])
+        self.assertEqual(
+            handler.log,
+            ['connection reset by peer'
+             ' (client closed "%s" before acknowledgement)' % handler.path])
 
     def test_handle_polling_handles_disappearing_files(self):
         handler = MyRequestHandlerForTests()
@@ -341,7 +368,7 @@ class TestMyRequestHandler(unittest.TestCase):
         self.assertEqual(handler.error_body,
                          "File not found: /nosuchfile.txt")
         self.assertEqual(handler.log,
-             ["[Errno 2] No such file or directory: 'nosuchfile.txt'"])
+                         ["[Errno 2] No such file or directory: 'nosuchfile.txt'"])
 
     def test_handle_command(self):
         handler = MyRequestHandlerForTests()
@@ -373,6 +400,21 @@ class TestMyRequestHandler(unittest.TestCase):
         self.assertTrue(b'cat: README.rst: no such file' in body,
                         body)
 
+    def test_handle_command_with_warnings(self):
+        handler = MyRequestHandlerForTests()
+        with patch('subprocess.Popen', PopenStub('hello', 'warning: blah blah', 0)):
+            body = handler.handle_command('python setup.py --long-description')
+        self.assertEqual(handler.status, 200)
+        self.assertEqual(handler.headers['Content-Type'],
+                         "text/html; charset=UTF-8")
+        self.assertEqual(handler.headers['Content-Length'],
+                         str(len(body)))
+        self.assertEqual(handler.headers['Cache-Control'],
+                         "no-cache, no-store, max-age=0")
+        self.assertFalse('X-Restview-Mtime' in handler.headers)
+        self.assertTrue(b'hello' in body, body)
+        self.assertTrue(b'blah blah' not in body, body)
+
     def test_handle_command_returns_error_with_watch_files(self):
         handler = MyRequestHandlerForTests()
         with patch('subprocess.Popen', PopenStub('', 'cat: README.rst: no such file', 1)):
@@ -421,9 +463,9 @@ class TestMyRequestHandler(unittest.TestCase):
         handler = MyRequestHandlerForTests()
         handler.collect_files = lambda dir: ['a.txt', 'b/c.txt']
         handler.render_dir_listing = lambda title, files: \
-                unicode("<title>%s</title>\n%s" % (
-                    title,
-                    '\n'.join('%s - %s' % (path, fn) for path, fn in files)))
+            unicode("<title>%s</title>\n%s" % (
+                title,
+                '\n'.join('%s - %s' % (path, fn) for path, fn in files)))
         body = handler.handle_dir('/path/to/dir')
         self.assertEqual(handler.status, 200)
         self.assertEqual(handler.headers['Content-Type'],
@@ -440,9 +482,9 @@ class TestMyRequestHandler(unittest.TestCase):
         handler = MyRequestHandlerForTests()
         handler.collect_files = lambda dir: ['a.txt', os.path.join('b', 'c.txt')]
         handler.render_dir_listing = lambda title, files: \
-                unicode("<title>%s</title>\n%s" % (
-                    title,
-                    '\n'.join('%s - %s' % (path, fn) for path, fn in files)))
+            unicode("<title>%s</title>\n%s" % (
+                title,
+                '\n'.join('%s - %s' % (path, fn) for path, fn in files)))
         with patch('os.path.isdir', lambda fn: fn == 'subdir'):
             body = handler.handle_list([os.path.normpath('/path/to/file.txt'),
                                         'subdir'])
@@ -455,8 +497,7 @@ class TestMyRequestHandler(unittest.TestCase):
                          b"<title>RST files</title>\n"
                          b"0/file.txt - #path#to#file.txt\n"
                          b"1/a.txt - subdir#a.txt\n"
-                         b"1/b/c.txt - subdir#b#c.txt"
-                            .replace(b"#", os.path.sep.encode()))
+                         b"1/b/c.txt - subdir#b#c.txt".replace(b"#", os.path.sep.encode()))
 
 
 def doctest_MyRequestHandler_render_dir_listing():
@@ -569,7 +610,7 @@ def doctest_RestViewer_rest_to_html_css_url():
         <head>
         <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
         ...
-        <title></title>
+        <title>...</title>
         <link rel="stylesheet" href="http://example.com/my.css" type="text/css" />
         <style type="text/css">
         ...
@@ -595,7 +636,7 @@ def doctest_RestViewer_rest_to_html_strict_and_error_handling():
 
         >>> viewer = RestViewer('.')
         >>> viewer.stylesheets = None
-        >>> viewer.strict = True
+        >>> viewer.halt_level = 2
         >>> print(viewer.rest_to_html(b'''
         ... Some text with an `error
         ... ''', mtime=1364808683).strip())
@@ -628,7 +669,7 @@ def doctest_RestViewer_rest_to_html_strict_and_error_handling():
         </body>
         </html>
 
-        >>> stderr_patcher.stop()
+        >>> _ = stderr_patcher.stop()
 
     """
 
@@ -646,16 +687,14 @@ def doctest_RestViewer_rest_to_html_pypi_strict_and_error_handling():
         ... Hello
         ... -----
         ...
-        ... .. image:: https://example.com/buildstatus.png
+        ... .. include:: /etc/passwd
         ...
-        ... Here is a `relative link <example.com>`__.
-        ...
-        ... ''').strip())
+        ... ''').strip().replace("&quot;", '"'))
         ... # doctest: +ELLIPSIS,+REPORT_NDIFF
         <!DOCTYPE html>
         <html>
         <head>
-        <title>TransformError</title>
+        <title>SystemMessage</title>
         <style type="text/css">
         pre.error {
             ...
@@ -663,22 +702,61 @@ def doctest_RestViewer_rest_to_html_pypi_strict_and_error_handling():
         </style>
         </head>
         <body>
-        <h1>TransformError</h1>
+        <h1>SystemMessage</h1>
         <pre class="error">
-        link scheme not allowed: example.com
+        &lt;string&gt;:5: (WARNING/2) "include" directive disabled.
         </pre>
         <pre>
+        <BLANKLINE>
         Hello
         -----
         <BLANKLINE>
-        .. image:: https://example.com/buildstatus.png
+        .. include:: /etc/passwd
+        <BLANKLINE>
         <BLANKLINE>
-        Here is a `relative link &lt;example.com&gt;`__.
         </pre>
         </body>
         </html>
 
-        >>> stderr_patcher.stop()
+        >>> _ = stderr_patcher.stop()
+
+    """
+
+
+def doctest_RestViewer_rest_to_html_pypi_strict():
+    """Test for RestViewer.rest_to_html in --pypi-strict mode
+
+        >>> stderr_patcher = patch('sys.stderr', StringIO())
+
+        >>> viewer = RestViewer('.')
+        >>> viewer.stylesheets = None
+        >>> viewer.pypi_strict = True
+        >>> print(viewer.rest_to_html(b'''
+        ... Hello
+        ... -----
+        ...
+        ... `This is fine <http://www.example.com>`__.
+        ...
+        ... ''').strip().replace("&quot;", '"'))
+        ... # doctest: +ELLIPSIS,+REPORT_NDIFF
+        <?xml version="1.0" encoding="utf-8" ?>
+        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
+        <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
+        <head>
+        <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
+        ...
+        <title>Hello</title>
+        <style type="text/css">
+        ...
+        </head>
+        <body>
+        <div class="document" id="hello">
+        <h1 class="title">Hello</h1>
+        <BLANKLINE>
+        <p><a href="http://www.example.com" rel="nofollow">This is fine</a>.</p>
+        </div>
+        </body>
+        </html>
 
     """
 
@@ -747,7 +825,118 @@ class TestRestViewer(unittest.TestCase):
         viewer = RestViewer('.')
         viewer.server = Mock()
         viewer.serve()
-        viewer.server.serve_forever.assert_called_once()
+        self.assertEqual(viewer.server.serve_forever.call_count, 1)
+
+    def test_rest_to_html_halt_level(self):
+        viewer = RestViewer('.')
+        viewer.halt_level = 2
+        html = viewer.rest_to_html(b'`Hello')
+        self.assertIn('<title>SystemMessage</title>', html)
+
+    def test_rest_to_html_report_level(self):
+        viewer = RestViewer('.')
+        viewer.report_level = 1
+        html = viewer.rest_to_html(b'.. _unused:\n\nEtc.')
+        self.assertIn('System Message: INFO/1', html)
+
+    @patch('readme_renderer.rst.clean', Mock(return_value=None))
+    def test_rest_to_html_pypi_strict_clean_failure(self):
+        # Certain versions of readme_renderer could return `None`
+        # from the clean() helper.  New versions don't (or at least
+        # do in different circumstances), so we have to mock out
+        # the helper to keep this test.
+        viewer = RestViewer('.')
+        viewer.pypi_strict = True
+        html = viewer.rest_to_html(b'''
+            [http://localhost:3000](http://localhost:3000)
+        ''')
+        self.assertIn('<title>ValueError</title>', html)
+        self.assertIn('Output cleaning failed', html)
+
+    def make_error(self, msg, source='file.rst', line=None,
+                   level=docutils.utils.Reporter.ERROR_LEVEL):
+        sm = docutils.nodes.system_message(
+            msg, level=level, type=docutils.utils.Reporter.levels[level],
+            source=source, line=line)
+        return docutils.utils.SystemMessage(sm, level)
+
+    def test_docutils_exception_messages(self):
+        err = self.make_error('dangling participle', 'file.rst', 42)
+        # Let's make sure docutils hasn't changed their error format.
+        self.assertEqual(str(err), 'file.rst:42: (ERROR/3) dangling participle')
+
+    def test_docutils_exception_messages_no_source_path(self):
+        err = self.make_error('dangling participle', None, 42)
+        # Let's make sure docutils hasn't changed their error format.
+        self.assertEqual(str(err), 'None:42: (ERROR/3) dangling participle')
+
+    def test_extract_line_info(self):
+        eli = RestViewer.extract_line_info
+        err = self.make_error('dangling participle', 'file.rst', 42)
+        self.assertEqual(eli(err, 'file.rst'), 42)
+
+    def test_extract_line_unknown_file(self):
+        eli = RestViewer.extract_line_info
+        err = self.make_error('dangling participle', None, 42)
+        self.assertEqual(eli(err, None), 42)
+
+    def test_extract_line_info_wrong_file(self):
+        eli = RestViewer.extract_line_info
+        err = self.make_error('dangling participle', 'file.rst', 42)
+        self.assertEqual(eli(err, 'unrelated.rst'), None)
+
+    def test_extract_line_info_other_kind_of_exception(self):
+        eli = RestViewer.extract_line_info
+        err = KeyboardInterrupt()
+        self.assertEqual(eli(err, 'file.rst'), None)
+
+    def test_highlight_line_no_line(self):
+        hl = RestViewer.highlight_line
+        source = 'line <1>\nline <2>\nline <3>\n'
+        self.assertEqual(hl(source, None),
+                         'line &lt;1&gt;\nline &lt;2&gt;\nline &lt;3&gt;\n')
+
+    def test_highlight_line_beyond_eof(self):
+        hl = RestViewer.highlight_line
+        source = 'line <1>\nline <2>\nline <3>\n'
+        self.assertEqual(hl(source, 42),
+                         'line &lt;1&gt;\nline &lt;2&gt;\nline &lt;3&gt;\n')
+
+    def test_highlight_line_before_bof(self):
+        hl = RestViewer.highlight_line
+        source = 'line <1>\nline <2>\nline <3>\n'
+        self.assertEqual(hl(source, 0),
+                         'line &lt;1&gt;\nline &lt;2&gt;\nline &lt;3&gt;\n')
+
+    def test_highlight_line_first(self):
+        hl = RestViewer.highlight_line
+        source = 'line <1>\nline <2>\nline <3>\n'
+        self.assertEqual(
+            hl(source, 1),
+            '<span class="highlight">line &lt;1&gt;\n</span>'
+            'line &lt;2&gt;\n'
+            'line &lt;3&gt;\n'
+        )
+
+    def test_highlight_line_middle(self):
+        hl = RestViewer.highlight_line
+        source = 'line <1>\nline <2>\nline <3>\n'
+        self.assertEqual(
+            hl(source, 2),
+            'line &lt;1&gt;\n'
+            '<span class="highlight">line &lt;2&gt;\n</span>'
+            'line &lt;3&gt;\n'
+        )
+
+    def test_highlight_line_last(self):
+        hl = RestViewer.highlight_line
+        source = 'line <1>\nline <2>\nline <3>\n'
+        self.assertEqual(
+            hl(source, 3),
+            'line &lt;1&gt;\n'
+            'line &lt;2&gt;\n'
+            '<span class="highlight">line &lt;3&gt;\n</span>'
+        )
 
 
 class TestGlobals(unittest.TestCase):
@@ -764,11 +953,7 @@ class TestGlobals(unittest.TestCase):
             Thread.assert_called_once_with(target=webbrowser.open,
                                            args=('http://example.com',))
             Thread.return_value.setDaemon.assert_called_once_with(True)
-            Thread.return_value.start.assert_called_once()
-
-    def test_trim_docstring(self):
-        from restview.pypi_support import trim_docstring
-        self.assertEqual(trim_docstring(""), "")
+            self.assertEqual(Thread.return_value.start.call_count, 1)
 
 
 class TestMain(unittest.TestCase):
@@ -800,26 +985,25 @@ class TestMain(unittest.TestCase):
                             if serve_called:
                                 self.assertTrue(self._serve_called)
                             if browser_launched:
-                                launch_browser.assert_called_once()
+                                self.assertEqual(launch_browser.call_count, 1)
                             return stdout.getvalue(), stderr.getvalue()
 
     def test_help(self):
         stdout, stderr = self.run_main('--help')
-        self.assertTrue('restview [options] filename-or-directory' in stdout,
-                        stdout)
+        self.assertTrue('restview [options] root' in stdout, stdout)
 
     def test_error_when_no_arguments(self):
         stdout, stderr = self.run_main(rc=2)
         self.assertEqual(stderr.splitlines()[-1],
-             'restview: error: at least one argument expected')
+                         'restview: error: at least one argument expected')
 
     def test_error_when_both_command_and_file_specified(self):
         stdout, stderr = self.run_main('-e', 'cat README.rst', 'CHANGES.rst',
                                        rc=2)
-        self.assertEqual(stderr.splitlines()[-1],
-             'restview: error: specify a command (-e) or a file/directory,'
-             ' but not both',
-             stderr)
+        self.assertEqual(
+            stderr.splitlines()[-1],
+            'restview: error: specify a command (-e) or a file/directory, but not both',
+            stderr)
 
     def test_all_is_well(self):
         self.run_main('.', serve_called=True, browser_launched=True)
@@ -836,12 +1020,18 @@ class TestMain(unittest.TestCase):
         with patch.object(RestViewer, 'listen'):
             with patch.object(RestViewer, 'close'):
                 self.run_main('-l', '0.0.0.0:8080', '.',
-                            serve_called=True, browser_launched=True)
+                              serve_called=True, browser_launched=False)
 
     def test_specify_invalid_listen_address(self):
         stdout, stderr = self.run_main('-l', 'nonsense', '.', rc=2)
         self.assertEqual(stderr.splitlines()[-1],
-             'restview: error: Invalid address: nonsense')
+                         'restview: error: Invalid address: nonsense')
+
+    def test_specify_allowed_hosts(self):
+        with patch.object(RestViewer, 'listen'):
+            with patch.object(RestViewer, 'close'):
+                self.run_main('--allowed-hosts', 'localhost,example.com', '.',
+                              serve_called=True, browser_launched=False)
 
     def test_custom_css_url(self):
         self.run_main('.', '--css', 'http://example.com/my.css',
@@ -854,10 +1044,7 @@ class TestMain(unittest.TestCase):
 
 def test_suite():
     return unittest.TestSuite([
-        unittest.makeSuite(TestMyRequestHandler),
-        unittest.makeSuite(TestRestViewer),
-        unittest.makeSuite(TestGlobals),
-        unittest.makeSuite(TestMain),
+        unittest.defaultTestLoader.loadTestsFromName(__name__),
         doctest.DocTestSuite(optionflags=doctest.ELLIPSIS | doctest.REPORT_NDIFF),
         doctest.DocTestSuite('restview.restviewhttp'),
     ])
